/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: solver.c,v 1.315 2009/11/08 02:26:56 zlb Exp $ */

#define NEED_GetVariableArgs

#include "phg.h"
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <math.h>

/* the workaround when the local size of the matrix is 0 on some processes.
 * Typical test case:
 *      mpirun -np 4 poisson -fem_type=P1 -mesh_file=tetra.dat -pre_refines=2
 */

/* list of solvers */
#define NSolvers	(sizeof(phgSolverList_)/sizeof(phgSolverList_[0]) - 1)

/* Solver names */
const char *phgSolverNames_[] = {
    "none",		/* dummy entry used by other interfaces (BLOPEX) */
    "pcg",
    "gmres",
    "preonly",
    "ams",
    "gather",
#if USE_HYPRE
    "hypre",
    "hamg",
#endif
#if USE_X9AMG
    "x9amg",
#endif
#if USE_PETSC
    "petsc",
#endif
#if USE_TRILINOS
    "trilinos",
#endif
#if USE_MUMPS
    "mumps",
#endif
#if USE_PASTIX
    "pastix",
#endif
#if USE_PARDISO
    "pardiso",
#endif
#if USE_SUPERLU
    "superlu",
#endif
#if USE_SPOOLES
    "spooles",
#endif
#if USE_SSPARSE
    "ssparse",
#endif
#if USE_SPC
    "spc",
#endif
#if USE_LASPACK
    "laspack",
#endif
    "builtin",
    NULL
};

/* Solvers list */
OEM_SOLVER *phgSolverList_[] = {
    NULL,		/* dummy entry reserved for other interfaces */
    SOLVER_PCG,
    SOLVER_GMRES,
    SOLVER_PreOnly,
    SOLVER_AMS,
    SOLVER_GATHER,
#if USE_HYPRE
    SOLVER_HYPRE,
    SOLVER_HAMG,
#endif
#if USE_X9AMG
    SOLVER_X9AMG,
#endif
#if USE_PETSC
    SOLVER_PETSC,
#endif
#if USE_TRILINOS
    SOLVER_TRILINOS,
#endif
#if USE_MUMPS
    SOLVER_MUMPS,
#endif
#if USE_PASTIX
    SOLVER_PASTIX,
#endif
#if USE_PARDISO
    SOLVER_PARDISO,
#endif
#if USE_SUPERLU
    SOLVER_SUPERLU,
#endif
#if USE_SPOOLES
    SOLVER_SPOOLES,
#endif
#if USE_SSPARSE
    SOLVER_SSPARSE,
#endif
#if USE_SPC
    SOLVER_SPC,
#endif
#if USE_LASPACK
    SOLVER_LASPACK,
#endif
    SOLVER_BUILTIN,
};
static int default_solver = 0;

/* Parameters */
static INT maxit = 10000;
static BOOLEAN warn_maxit = TRUE;
static FLOAT rtol = 1e-10;
static FLOAT atol_ = 0.;	/* avoid conflicts with stdlib.h */
static FLOAT btol = 1e-10;

int
phgSolverRegisterOptions(void)
{
    int i;

    phgOptionsRegisterTitle("\nLinear solver options:", "\n", "solver");
    phgOptionsRegisterKeyword("-solver", "Default solver",
			      phgSolverNames, &default_solver);
    phgOptionsRegisterInt("-solver_maxit",
			  "Default maxi. number of iterations", &maxit);
    phgOptionsRegisterNoArg("-solver_warn_maxit",
			  "Warn if maxi. number of iterations attained",
				&warn_maxit);
    phgOptionsRegisterFloat("-solver_rtol", "Relative convergence tolerance "
			    "(||residual|| <= ||residual0|| * rtol)", &rtol);
    phgOptionsRegisterFloat("-solver_btol", "B convergence tolerance "
			    "(||residual|| <= ||RHS|| * btol)", &btol);
    phgOptionsRegisterFloat("-solver_atol", "Absolute convergence tolerance "
			    "(||residual|| <= atol)", &atol_);

    /* Register eigensolver options */
    phgEigenSolve(NULL, NULL, 0, 0, 0.0, NULL, NULL, NULL);

    for (i = 0; i < NSolvers; i++)
	if (phgSolverList[i] != NULL &&
	    phgSolverList[i]->RegisterOptions != NULL)
	    phgSolverList[i]->RegisterOptions();

    return 0;
}

static int
send_matrix_to_oem_solver(SOLVER *solver)
/* sends data to OEM solver */
{
    MAT *mat = solver->mat;
    INT i, j;
    const MAT_ROW *row;
    FLOAT total_cols;

    FunctionEntry;

    if (solver->sent_to_oem)
	Return 0;

    if (!solver->oem_created) {
	solver->oem_created = TRUE;
	solver->oem_solver->Create(solver);
    }

    /* check whether the solver has AddMatrixEntries function */
    if (solver->oem_solver->AddMatrixEntries == NULL)
	Return 0;

    if (mat->type == PHG_MATRIX_FREE && mat->blocks == NULL)
	Return 0;

    assert(mat->type != PHG_DESTROYED);

    phgInfo(1, "sending matrix to OEM solver...\n");

    /* send matrix to OEM solver and free PHG's matrix */
    j = mat->rmap->partition[mat->rmap->rank];
    total_cols = 0.0;
    for (i = 0; i < mat->rmap->nlocal; i++, j++) {
	row = phgMatGetRow(mat, i);
	if (row->ncols <= 0)
	    phgError(1, "%s: matrix row %d is empty!\n", __func__, i);
	total_cols += row->ncols;
	solver->oem_solver->AddMatrixEntries(solver, 1, &j, row->ncols,
					     row->cols, row->data);

	if (mat->blocks == NULL && mat->refcount == 0 && mat->rows != NULL
		/*&& (solver->pc == NULL || solver->pc->solver != solver)*/) {
	    /* free PHG matrix row */
	    phgFree(mat->rows[i].cols);
	    phgFree(mat->rows[i].data);
	    mat->rows[i].cols = NULL;
	    mat->rows[i].data = NULL;
	    mat->rows[i].ncols = mat->rows[i].alloc = 0;
	}
    }

    phgInfo(2, "average cols = %lg\n",
		((double)total_cols) / solver->mat->rmap->nlocal);

    if (mat->refcount == 0
	/*&& (solver->pc == NULL || solver->pc->solver != solver)*/)
	    phgMatFreeMatrix(mat);

    solver->sent_to_oem = TRUE;

    Return 0;
    return 0;		/* to avoid possible gcc warnings */
}

BOOLEAN
phgSolverDenseLU(int n, FLOAT *a0, int pvt[])
/* LU factorization (partial pivoting). TODO: call LAPACK for large n */
{
    int i, j, k;
    FLOAT d, r;
#define a(i,j)	(a0[(i) * n + (j)])

    for (i = 0; i < n; i++) {
	d = Fabs(a(i,i));
	k = i;
	for (j = i + 1; j < n; j++) {
	    if ((r = Fabs(a(j,i))) > d) {
		d = r;
		k = j;
	    }
	}
	if (d == 0.0) {
#if 0 * DEBUG
	    printf("Dense solver failed: i = %d\n", i);
	    for (j = 0; j < n; j++) {
		for (k = 0; k < n; k++)
		    printf(" %9.2le", (k >= i || k >= j) ? (double)a(j,k) : 0.);
	    }
#endif
	    return FALSE;
	}
	pvt[i] = k;
	if (k != i) {
	    /* exchange row i and row k */
	    for (j = i; j < n; j++) {
		d = a(i,j);
		a(i,j) = a(k,j);
		a(k,j) = d;
	    }
	}
	if ((d = a(i,i)) != 1.0) {
	    a(i,i) = d = 1.0 / d;
	    for (j = i + 1; j < n; j++)
		a(i,j) *= d;
	}
	for (j = i + 1; j < n; j++) {
	    if ((d = a(j,i)) == 0.0)
		continue;
	    for (k = i + 1; k < n; k++)
		a(j,k) -= d * a(i,k);
	}
    }
#undef a

    return TRUE;
}

void
phgSolverDenseSV(int n, FLOAT *a0, int pvt[], int m, FLOAT *b0)
/* solves L*U*X = B. TODO: call LAPACK for large n */
{
    int i, j, k;
    FLOAT d;
#define a(i,j)	(a0[(i) * n + (j)])
#define b(i,j)	(b0[(i) * m + (j)])

    for (i = 0; i < n; i++) {
	k = pvt[i];
	if (k != i) {
	    /* exchange row i with row k */
	    for (j = 0; j < m; j++) {
		d = b(i,j);
		b(i,j) = b(k,j);
		b(k,j) = d;
	    }
	}
	if ((d = a(i,i)) != 1.0) {
	    for (j = 0; j < m; j++)
		b(i,j) *= d;
	}
	for (j = i + 1; j < n; j++) {
	    if ((d = a(j,i)) == 0.0)
		continue;
	    for (k = 0; k < m; k++)
		b(j,k) -= d * b(i,k);
	}
    }

    for (i = n - 2; i >= 0; i--)
	for (j = i + 1; j < n; j++) {
	    d = a(i,j);
	    for (k = 0; k < m; k++)
		b(i,k) -= d * b(j,k);
	}
#undef a
#undef b

    return;
}

BOOLEAN
phgSolverDenseSolver(int n, int m, FLOAT *a0, FLOAT *b0)
/* solves AX = B, returns TRUE if successful and FALSE if A is singular.
 * TODO: call LAPACK for large n */
{
#if 1
    int i, j, k;
    FLOAT d, r;
#define a(i,j)	(a0[(i) * n + (j)])
#define b(i,j)	(b0[(i) * m + (j)])

    for (i = 0; i < n; i++) {
	d = Fabs(a(i,i));
	k = i;
	for (j = i + 1; j < n; j++) {
	    if ((r = Fabs(a(j,i))) > d) {
		d = r;
		k = j;
	    }
	}
	if (d == 0.0) {
#if 0 * DEBUG
	    printf("Dense solver failed: i = %d\n", i);
	    for (j = 0; j < n; j++) {
		for (k = 0; k < n; k++)
		    printf(" %9.2le", (k >= i || k >= j) ? (double)a(j,k) : 0.);
		for (k = 0; k < m; k++)
		    printf(" = %9.2le\n", (double)b(j,k));
	    }
#endif
	    return FALSE;
	}
	if (k != i) {
	    /* exchange row i with row k */
	    for (j = i; j < n; j++) {
		d = a(i,j);
		a(i,j) = a(k,j);
		a(k,j) = d;
	    }
	    for (j = 0; j < m; j++) {
		d = b(i,j);
		b(i,j) = b(k,j);
		b(k,j) = d;
	    }
	}
	if ((d = a(i,i)) != 1.0) {
	    d = 1.0 / d;
	    for (j = i + 1; j < n; j++)
		a(i,j) *= d;
	    for (j = 0; j < m; j++)
		b(i,j) *= d;
	}
	for (j = i + 1; j < n; j++) {
	    if ((d = a(j,i)) == 0.0)
		continue;
	    for (k = i + 1; k < n; k++)
		a(j,k) -= d * a(i,k);
	    for (k = 0; k < m; k++)
		b(j,k) -= d * b(i,k);
	}
    }

    for (i = n - 2; i >= 0; i--)
	for (j = i + 1; j < n; j++) {
	    d = a(i,j);
	    for (k = 0; k < m; k++)
		b(i,k) -= d * b(j,k);
	}
#undef a
#undef b
#else
    int *pvt = phgAlloc(n * sizeof(*pvt));

    if (phgSolverDenseLU(n, a0, pvt) == FALSE)
	return FALSE;
    phgSolverDenseSV(n, a0, pvt, m, b0);
    phgFree(pvt);
#endif

    return TRUE;
}

static void
update_rhs(MAT *smat, VEC *srhs, MAT *dmat, VEC *drhs, FLOAT coeff,
	   BOOLEAN diag)
/* This function computes:
 *	drhs -= coeff * B(dmat) * D(smat)^(-1) * srhs
 * where:
 *	A(mat) = mat with all bdry columns set to 0,
 *	B(mat) = the matrix composed of the bdry columns of mat,
 *	D(mat) = diag matrix with non-bdry part = I and bdry part = diag(mat)
 *
 * Note: if mat->handle_bdry_eqns == TRUE, then after MAT assembly, only the
 * diagonal entries in the bdry rows can be nonzero, and A is stored in
 * mat->rows, B+D is stored in mat->bdry_rows (note both mat->rows and
 * mat->bdry_rows contain D). Since mat->rows might have been destroyed when
 * sending data to OEM solvers, we get D from mat->bdry_rows.
 *
 * TODO: replace bdry_rows with a MAT object, and use mat-vec operations. */
{
    MAT *mat;
    VEC *rhs;
    MAT_ROW *row;
    INT i, j, k;
#if USE_MPI
    INT O2Gsize, n, *rlist, *slist;
    FLOAT *rbuff, *sbuff;
    BYTE *flags;
    int rank, rsize, *sdsps, *scnts, *rdsps, *rcnts;
#endif	/* USE_MPI */

    if (!smat->handle_bdry_eqns)
	return;

    assert(srhs->nvec == drhs->nvec);

    if (smat->blocks != NULL) {
	FLOAT *sdata, *ddata, c0, c;
	INT p = smat->blocks->p, q = smat->blocks->q;
	assert(smat->type == PHG_MATRIX_FREE || smat->type == PHG_DESTROYED);
	assert(dmat->blocks != NULL && p == q);
	if (p == 1) {
	    assert(dmat->blocks->p == 1 && dmat->blocks->q == 1);
	    if ((c = dmat->blocks->coeff[0]) == 0.)
		return;
	    c0 = smat->blocks->coeff[0];
	    assert(c0 != 0.);
	    update_rhs(smat->blocks->pmat[0], srhs,
		       dmat->blocks->pmat[0], drhs, c / c0, TRUE);
	    return;
	}
	sdata = srhs->data;
	for (j = 0; j < q; j++) {
	    VEC *rhs0;
	    MAT *mat0;
	    /* update column j */
	    mat0 = smat->blocks->pmat[j * q + j];	/* diagonal block */
	    if (mat0 == NULL) {
		/* zero diagonal, block which shouldn't have bdry entries */
		sdata += srhs->nvec * smat->blocks->ncols[j];
		continue;
	    }
	    assert(mat0 != NULL && mat0->type != PHG_MATRIX_FREE);
	    rhs0 = phgMapCreateVec(mat0->cmap, srhs->nvec);
	    phgFree(rhs0->data);
	    rhs0->data = sdata;
	    ddata = drhs->data;
	    c0 = smat->blocks->coeff[j * q + j];
	    assert(c0 != 0.);
	    c0 = 1. / c0;
	    for (i = 0; i < p; i++) {
		c = dmat->blocks->coeff[i * q + j];
		mat = dmat->blocks->pmat[i * q + j];
		if (c != 0. && mat != NULL) {
		    assert(mat->type != PHG_MATRIX_FREE);
		    rhs = phgMapCreateVec(mat->rmap, drhs->nvec);
		    phgFree(rhs->data);
		    rhs->data = ddata;
		    assert((mat->handle_bdry_eqns == TRUE) == (i == j));
		    mat->handle_bdry_eqns = TRUE;
		    update_rhs(mat0, rhs0, mat, rhs, c * c0, i == j);
		    mat->handle_bdry_eqns = (i == j);
		    rhs->data = NULL;
		    phgVecDestroy(&rhs);
		}
		ddata += drhs->nvec * smat->blocks->nrows[i];
	    }
	    rhs0->data = NULL;
	    phgVecDestroy(&rhs0);
	    sdata += srhs->nvec * smat->blocks->ncols[j];
	}

	return;
    }

    assert(smat->assembled && dmat->assembled);
    assert(dmat->handle_bdry_eqns);
    /* The diagonal block must be square */
    assert(smat->rmap->nlocal == smat->cmap->nlocal);
    /* The number of columns in the src and dest blocks must match */
    assert(smat->cmap->nlocal == dmat->cmap->nlocal);

    if (!srhs->assembled)
	phgVecAssemble(srhs);

    if (!drhs->assembled)
	phgVecAssemble(drhs);

    mat = dmat;
    rhs = drhs;

#if USE_MPI
    /* prepare communications for getting off-proc bdry entries */
    O2Gsize = mat->localsize - mat->cmap->nlocal;
    rsize = mat->bdry_localsize - mat->cmap->bdry_nlocal;
    rlist = phgAlloc(rsize * sizeof(*rlist));
    rbuff = phgAlloc(rsize * sizeof(*rbuff));
    flags = phgCalloc(O2Gsize, sizeof(*flags));
    scnts = phgAlloc(4 * mat->cmap->nprocs * sizeof(*scnts));
    sdsps = scnts + mat->cmap->nprocs;
    rcnts = sdsps + mat->cmap->nprocs;
    rdsps = rcnts + mat->cmap->nprocs;

    n = mat->cmap->partition[mat->cmap->rank];

    /* expand off-proc bdry entries for faster accessing */
    for (i = mat->cmap->bdry_nlocal; i < mat->bdry_localsize; i++) {
	assert(mat->bdry[i] >= mat->cmap->nlocal);
	flags[mat->bdry[i] - mat->cmap->nlocal] = 1;
    }

    rank = 0;
    bzero(rcnts, mat->cmap->nprocs * sizeof(*rcnts));
    for (i = 0, j = 0; i < O2Gsize; i++) {
	k = mat->ordering[i];
	if (flags[k] == 0)
	    continue;
	k = rlist[j++] = mat->O2Gmap[k];
	assert(j == 1 || rlist[j - 1] > rlist[j - 2]);
	while (rank < mat->cmap->nprocs &&
	       mat->cmap->partition[rank + 1] <= k)
	    rank++;
	assert(rank < mat->cmap->nprocs && rank != mat->cmap->rank);
	rcnts[rank]++;
    }
    assert(j == rsize);
    MPI_Alltoall(rcnts, 1, MPI_INT, scnts, 1, MPI_INT, mat->cmap->comm);
    i = j = 0;
    for (rank = 0; rank < mat->cmap->nprocs; rank++) {
	rdsps[rank] = i;
	sdsps[rank] = j;
	i += rcnts[rank];
	j += scnts[rank];
    }
    /* Note: i/j = total number of entries to recv/send */
    assert(i == rsize);
    slist = phgAlloc(j * sizeof(*slist));
    sbuff = phgAlloc(j * sizeof(*sbuff));
    phgAlltoallv(rlist, rcnts, rdsps, PHG_MPI_INT,
		 slist, scnts, sdsps, PHG_MPI_INT, mat->cmap->comm);
    /* collect entries to send */
    for (i = 0; i < j; i++) {
	k = slist[i] - n;
	assert(k >= 0 && k < smat->cmap->nlocal);
#if DEBUG
	/* Note: k must be a boundary entry, otherwise it means incompatibility
	 * of boundary masks in different submeshes */
	if (mat->cmap->bdry_nlocal == 0 ||
		bsearch(&k, mat->cmap->bdry, mat->cmap->bdry_nlocal, sizeof(k),
			phgCompINT) == NULL) {
	    /* identify the source process */
	    j = 0;
	    while (j + 1 < mat->cmap->nprocs && sdsps[j + 1] < i)
		j++;
	    phgError(1, "%s:%d, row %d from p%d (%s) is not a bdry row\n",
			__FILE__, __LINE__, k + n, j, phgProcessorNames[j]);
	}
	if (smat->bdry_rows[k].ncols != 1 ||
	    smat->bdry_rows[k].cols[0] != k)
	    phgError(1, "%s:%d, unexpected error: row %d, ncols = %d\n",
			__FILE__, __LINE__, k, smat->bdry_rows[k].ncols);
#endif	/* DEBUG */
	sbuff[i] = srhs->data[k] / smat->bdry_rows[k].data[0];
    }
    phgFree(rlist);
    phgFree(slist);
    phgAlltoallv(sbuff, scnts, sdsps, PHG_MPI_FLOAT,
		 rbuff, rcnts, rdsps, PHG_MPI_FLOAT, mat->cmap->comm);
    phgFree(sbuff);
    phgFree(scnts);
    /* save received RHS's in srhs->offp_data */
    phgFree(srhs->offp_data);
    srhs->offp_data = phgAlloc(O2Gsize * sizeof(*srhs->offp_data));
    for (i = 0, j = 0; i < O2Gsize; i++) {
	k = mat->ordering[i];
	if (flags[k] == 0)
	    continue;
	srhs->offp_data[k] = rbuff[j++];
    }
    phgFree(rbuff);
#endif	/* USE_MPI */

    for (i = 0, row = mat->bdry_rows; i < mat->rmap->nlocal; i++, row++) {
	for (j = 0; j < row->ncols; j++) {
	    if ((k = row->cols[j]) == i && diag)
		continue;		/* skip column on diagonal */
#if USE_MPI
	    assert(k < mat->cmap->nlocal || mat->cmap->nprocs > 1);
	    if (k >= mat->cmap->nlocal) {
		k -= mat->cmap->nlocal;
		assert(flags[k] != 0);
		rhs->data[i] -= coeff * srhs->offp_data[k] * row->data[j];
		continue;
	    }
#endif	/* USE_MPI */
	    assert(smat->bdry_rows[k].ncols == 1 &&
		   smat->bdry_rows[k].cols[0] == k);
	    rhs->data[i] -= coeff * srhs->data[k] * row->data[j] /
					smat->bdry_rows[k].data[0];
	}
    }
#if USE_MPI
    phgFree(flags);
    if (O2Gsize > 0)
	bzero(srhs->offp_data, O2Gsize * sizeof(*srhs->offp_data));
#endif	/* USE_MPI */
}

void
phgSolverUpdateRHS(SOLVER *solver)
{
    if (solver->rhs_updated)
	return;
    if (!solver->sent_to_oem && !solver->mat->assembled)
	phgMatAssemble(solver->mat);
    update_rhs(solver->mat, solver->rhs, solver->mat, solver->rhs, 1.0, TRUE);
    solver->rhs_updated = TRUE;
}

int
phgSolverResetRHS(SOLVER *solver)
{
    assert(solver->rhs != NULL || solver->mat != NULL);
    if (solver->rhs == NULL) {
	solver->rhs = phgMapCreateVec(solver->mat->rmap, 1);
    }
    else {
	bzero(solver->rhs->data, solver->rhs->nvec *
			solver->rhs->map->nlocal * sizeof(*solver->rhs->data));
#if USE_MPI
	if (solver->rhs->localsize > solver->rhs->map->nlocal)
	    bzero(solver->rhs->offp_data, sizeof(*solver->rhs->offp_data) *
			(solver->rhs->localsize - solver->rhs->map->nlocal));
#endif	/* USE_MPI */
    }
    phgVecDisassemble(solver->rhs);
    solver->rhs_updated = FALSE;
    return 0;
}

/*-------------------------- User interface --------------------------------*/

int
phgSolverInitialize(int *argc, char ***argv)
{
    int i = 0;

    for (i = 0; i < NSolvers; i++)
	if (phgSolverList[i] != NULL && phgSolverList[i]->Initialize != NULL)
	    phgSolverList[i]->Initialize(argc, argv);

    return 0;
}

int
phgSolverFinalize(void)
{
    int i = 0;

    for (i = 0; i < NSolvers; i++)
	if (phgSolverList[i] != NULL && phgSolverList[i]->Finalize != NULL)
	    phgSolverList[i]->Finalize();

    return 0;
}

MAT *
phgSolverGetMat(SOLVER *solver)
{
    solver->mat->refcount++;
    return solver->mat;
}

SOLVER *
phgSolverMat2Solver(OEM_SOLVER *oem_solver, MAT *mat)
{
    SOLVER *solver = phgCalloc(1, sizeof(*solver));
    int j;

    solver->oem_solver = oem_solver;
    if (solver->oem_solver == NULL && default_solver >= 0) {
	static BOOLEAN warned = FALSE;
	solver->oem_solver = phgSolverList[default_solver];
	if (!warned && mat->rmap->rank == 0 && solver->oem_solver == NULL) {
	    warned = TRUE;
	    phgWarning("solver \"%s\" not available, will use the first "
			"available solver.\n", phgSolverNames[default_solver]);
	}
    }

    /* fall back to first available solver if the requested solver is not
     * available or is SOLVER_DEFAULT */
    if (solver->oem_solver == NULL) {
	for (j = 0; j < NSolvers; j++)
	    if (phgSolverList[j] != NULL) {
		solver->oem_solver = phgSolverList[j];
		break;
	    }
    }

    if (solver->oem_solver == NULL) {
	phgInfo(-1, "phgSolverCreate: no solver available\n");
	phgError(1, "phgSolverCreate: abort.\n");
    }

    if (solver->oem_solver->parallel == FALSE && mat->rmap->nprocs > 1)
	phgError(1, "solver \"%s\" can't be used for parallel systems.\n",
					solver->oem_solver->name);

    phgInfo(1, "using linear solver \"%s\".\n", solver->oem_solver->name);

    mat->refcount++;
    solver->mat = mat;
    solver->rhs = phgMapCreateVec(mat->rmap, 1);
    solver->rhs->mat = mat;	/* reverse pointer to the matrix */
    solver->sent_to_oem = FALSE;
    solver->oem_created = FALSE;
    solver->oem_data = NULL;

    solver->rtol = rtol;
    solver->atol = atol_;
    solver->btol = btol;
    solver->maxit = maxit;
    solver->warn_maxit = warn_maxit;
    solver->pc = NULL;

    if (solver->oem_solver->Init != NULL)
	solver->oem_solver->Init(solver);

    return solver;
}

SOLVER *
phgSolverCreate(OEM_SOLVER *oem_solver, DOF *u, ...)
{
    MAP *map;
    MAT *mat;
    int ndof;
    DOF **dofs;
    SOLVER *solver;

    GetVariableArgs(ndof, dofs, u, DOF *);

    map = phgMapCreateN(ndof, dofs);
    phgFree(dofs);
    mat = phgMapCreateMat(map, map);
    phgMapDestroy(&map);

    mat->handle_bdry_eqns = TRUE;

    solver = phgSolverMat2Solver(oem_solver, mat);
    phgMatDestroy(&mat);
    phgVecDisassemble(solver->rhs);
    solver->rhs_updated = FALSE;

    return solver;
}

int
phgSolverDestroy(SOLVER **solver_ptr)
{
    SOLVER *solver = *solver_ptr;

    if (solver == NULL)
	return 0;

    solver->oem_solver->Destroy(solver);

    phgMatDestroy(&solver->mat);

    phgVecDestroy(&solver->rhs);

    if (solver->pc != NULL) {
	if (solver->pc->solver != NULL && solver->pc->owns_solver)
	    phgSolverDestroy(&solver->pc->solver);
	phgFree(solver->pc);
    }

    phgFree(solver);

    *solver_ptr = NULL;

    return 0;
}

int
phgSolverAssemble(SOLVER *solver)
{
    FunctionEntry;

    assert(solver->mat != NULL);

    if (!solver->mat->assembled)
	phgMatAssemble(solver->mat);

    if (!solver->sent_to_oem)
	send_matrix_to_oem_solver(solver);

    Return (solver->oem_solver->Assemble == NULL ? 0 :
		solver->oem_solver->Assemble(solver));
}

int
phgSolverVecSolve(SOLVER *solver, BOOLEAN destroy, VEC *x)
/* returns number of iterations, < 0 ==> error */
{
    INT i, j;
    FLOAT *rhs;
    size_t mem_peak = 0;

    assert(solver->rhs != NULL && solver->mat != NULL);

    if (solver->mat->cmap->nlocal != x->map->nlocal) {
	phgError(1, "(%s:%d) inconsistent RHS!\n", __FILE__, __LINE__);
    }

    if (solver->mat->handle_bdry_eqns == TRUE &&
	   solver->rhs->mat != solver->mat)
	phgError(1, "(%s:%d) solver->rhs->mat != solver->mat.\n",
			__FILE__, __LINE__);

    phgSolverAssemble(solver);

    if (!solver->rhs->assembled)
	phgVecAssemble(solver->rhs);

    if (!solver->rhs_updated)
	phgSolverUpdateRHS(solver);

    /* send RHS to OEM solver and free PHG's RHS */
    if (solver->oem_solver->AddRHSEntries != NULL) {
	j = solver->mat->rmap->partition[solver->mat->rmap->rank];
	rhs = solver->rhs->data;
	for (i = 0; i < solver->mat->rmap->nlocal; i++, j++)
	    solver->oem_solver->AddRHSEntries(solver, 1, &j, rhs++);
    }

    if (phgVerbosity >= 1) {
	mem_peak = phgMemoryPeakReset(0);
	phgInfo(0, "calling OEM solver...\n");
    }
    i = solver->oem_solver->Solve(solver, x, destroy);
    if (phgVerbosity >= 1) {
	mem_peak = phgMemoryPeakRestore(mem_peak);
	phgInfo(0, "%0.4lgMB memory used in OEM solver.\n",
		mem_peak / (1024.0 * 1024.0));
    }

    if (x->map->rank == 0) {
	if (i < 0)
	    phgError(1, "error occured in OEM solver \"%s\".\n",
		     solver->oem_solver->name);
	if (solver->warn_maxit && i >= solver->maxit && solver->rtol > 0.0)
	    phgWarning("maxit attained in OEM solver, "
		       "the result may be inaccurate.\n");
    }

    return i;
}

int
phgSolverSolve(SOLVER *solver, BOOLEAN destroy, DOF *u, ...)
/* returns number of iterations, < 0 ==> error */
{
    int i, ndof;
    DOF **dofs;
    VEC *x;

    GetVariableArgs(ndof, dofs, u, DOF *);

    /* make sure the DOFs match those used in creating the solver */
    assert(solver->rhs->map->dofs != NULL);
    assert(solver->rhs->map->ndof == ndof);
    for (i = 0; i < ndof; i++) {
	if (dofs[i]->type != solver->rhs->map->dofs[i]->type)
	    phgError(1, "phgSolveSolve: DOF %d (%s), types mismatch: %s<=>%s\n",
		     i, dofs[i]->name, dofs[i]->type,
		     solver->rhs->map->dofs[i]->type);
	if (dofs[i]->dim != solver->rhs->map->dofs[i]->dim)
	    phgError(1, "phgSolveSolve: DOF %d (%s), dims mismatch: %d<=>%d\n",
		     i, dofs[i]->name, dofs[i]->dim,
		     solver->rhs->map->dofs[i]->dim);
    }

    x = phgMapCreateVec(solver->rhs->map, 1);
    /* copy initial solution to x */
    phgMapDofToLocalData(x->map, ndof, dofs, x->data);
    /* call solver */
    i = phgSolverVecSolve(solver, destroy, x);
    /* copy solution from x */
    phgMapLocalDataToDof(x->map, ndof, dofs, x->data);
    phgFree(dofs);
    phgVecDestroy(&x);

    return i;
}

/* The functions below are maintained only for backward compatibility */

void
phgSolverSetTol(SOLVER *solver, FLOAT user_tol)
{
#if USE_MPI
    FLOAT tol = user_tol;

    MPI_Bcast(&tol, 1, PHG_MPI_FLOAT, 0, solver->mat->rmap->comm);
    user_tol = tol;
#endif
    solver->rtol = user_tol;
}

void
phgSolverSetMaxIt(SOLVER *solver, INT user_maxit)
{
#if USE_MPI
    int maxit = user_maxit;

    MPI_Bcast(&maxit, 1, MPI_INT, 0, solver->mat->rmap->comm);
    user_maxit = maxit;
#endif
    solver->maxit = user_maxit;
}
